/////////////////////////////////////////////////////////////////////////
///   Arduino Stepper Control Demo
///   for Versatile Stepper Control
///    Will Cooke
/////////////////////////////////////////////////////////////////////////

// Set this to the total number of motors to control
#define NUM_MOTORS 4 

// Declare the additional functions defined
void stepper_initialize();
void stepper_setSteps(uint16_t mot, uint16_t steps, int16_t siz);
void stepper_stop(uint16_t mot);
void stepper_startAll();
void stepper_stopAll();
uint16_t stepper_stepsRemaining(uint16_t mot);

// These numbers are just for the demo to add to step size/speed
// for each motor
int16_t start0 = -2000;    // Starting speed of motor 0
int16_t start1 = -4000;    // Starting speed of motor 1
int16_t start2 = 3000;     // Starting speed of motor 2
int16_t inc0 = 50;         // Speed increment for motor 0
int16_t inc1 = 75;         // Speed increment for motor 1
int16_t inc2 = -90;        // Speed increment for motor 2

//////////////////////////////////////////////////////////////////////
///  In the Setup() function we set up serial communication with the 
///  PC so we can monitor what is happening.  Then we initialize the 
///  stepper functions.  That also sets up timer 0 compare A register 
///  to create the needed interrupts.  Then we set some initial motions 
///  for the steppers and start them moving.
///////////////////////////////////////////////////////////////////////

void setup() 
{
  Serial.begin(115200);              // Start talking to the PC.
  Serial.println("Hello, stepper!"); // And say hello.
  
  stepper_initialize();              // Initialize our stepper control.
  Serial.println("Ready...");        // And tell us when finished.
  stepper_setSteps(0,200, start0);     // Motor 0 200 steps
  stepper_setSteps(1, 400, start1);    // Motor 1 400 steps
  stepper_setSteps(2, 500, start2);    // Motor 2 500 steps
  stepper_startAll();                // Then start them.

}




///////////////////////////////////////////////////////////////////////
///  In the loop() function we can do about anything we like.  The 
///  steppers will be running in the background while our normal 
///  processing continue.  We can set, adjust, and monitor what the 
///  steppers are doing as well as do other, unrelated processing.
///  But remember that if you read OR write any of the stepper control 
///  variables directly you need to disable interrupts (noInterrupts() )
///  beforehand and re-enable ( interrupts() ) after you finish.  Don't 
///  leave the interrupts disabled for long or it will interfere with 
///  the motors running.
///  The interface functions below already disable and enable the 
///  interrupts so if you call them you don't need to worry about it. 
///  When you write your own interface functions you can use the ones 
///  below as examples of how to proceed.
///////////////////////////////////////////////////////////////////////

void loop() 
{
  delay(1000);    // A brief pause just for demonstration
  Serial.print(millis());
  Serial.println(" We are doing this while steppers step!");

  if(stepper_stepsRemaining(0) == 0)
  {
    start0 += inc0;
    if(start0 > 4000)
    {
      start0 = -2000;
    }
    stepper_setSteps(0,200, start0);
  }

  if(stepper_stepsRemaining(1) == 0)
  {
    start1 += inc1;
    if(start1 > 4000)
    {
      start1 = -3000;
    }
    stepper_setSteps(1, 300, start1);
  }

  if(stepper_stepsRemaining(2) == 0)
  {
    start2 += inc2;
    if(start2 < -4000)
    {
      start2 = 2000;
    }
    stepper_setSteps(2, 400, start2);
  }

  stepper_startAll();   // Restart the motors if they have stopped
  
}


//////////////////   Stepper code here //////////////////////////
//  The above code is mostly for demonstartion.  The code from 
//  here down are the functions to interface with the ISR and the 
//  ISR itself.  It can be used mostly as-is in your application. 
//  However, the code that interfaces to your motors to output 
//  the current pattern will need to be adapted to your hardware
//////////////////////////////////////////////////////////////////

// Half-wave step table
static const uint8_t stepper_halfWaveTable[] = { 1, 2, 4, 8 };

// Full-wave step table
static const uint8_t stepper_fullWaveTable[] = { 3, 6, 12, 9};

// Half-step step table
static const uint8_t stepper_halfStepTable[] = { 1, 3, 2, 6, 4, 12, 8, 9}; 


//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// 
//            EDIT HERE for your desired pattern
//
//////////////////////////////////////////////////////////////
// This pointer variable selects which table we want to use.  
// I have selected half wave table here.  Each table also has 
// a corresponding constant "stepper_indexShift" to shift the 
// phase accumulator the proper number of bits.  To select a 
// pattern uncomment the two lines you need and comment out 
// the others.

static const uint8_t* stepper_stepTable = stepper_halfWaveTable; 
static const uint8_t stepper_indexShift = 14;

//static const uint8_t* stepper_stepTable = stepper_fullWaveTable;
//static const uint8_t stepper_indexShift = 14;

//static const uint8_t* stepper_stepTable = stepper_halfStepTable;
//static const uint8_t stepper_indexShift = 13;
 

//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////
///  The variables below are part of the motor control.  The array 
///  "stepper_phase" is the phase accumulators for all the motors. 
///  The array "stepper_stepSize" holds the phase increment to be 
///  added to the accumulator at each interrupt.  
///  The array "stepper_steps" is the number of steps remaining for 
///  that motor.
///  The variable "stepper_running" indicates if the motors are 
///  currently running.
///
///  When variables are used inside an ISR as well as outside it, 
///  they should be declared "volatile" as they are here so the 
///  compiler knows that it may  change in unexpected ways (via the 
///  interrupt service routine.).  The compiler then knows that it 
///  must limit optimizations that it might otherwise make.
///////////////////////////////////////////////////////////////////

static volatile uint16_t stepper_phase[NUM_MOTORS];     // Phase Accumulators.
static volatile int16_t stepper_stepSize[NUM_MOTORS];   // Amount to step each time.
static volatile uint16_t stepper_steps[NUM_MOTORS];     // How many steps left?
static volatile bool stepper_running = false;           // Any motors running?


////////////////////////////////////////////////////////////////
///  Interface Functions
///  Below are the functions that allow you to control the 
///  stepper motors.  The ISR provides the actual stepping of 
///  the motors and these functions allow your program to set 
///  the step rate, the number of steps, and start and stop 
///  the motors.
////////////////////////////////////////////////////////////////
///  stepper_initialize
///  This function sets up things that need to be taken care of 
///  before the steppers can run properly.  The first thing it 
///  does is clear the ports being used for motor control.  
///  If you aren't sure how that works, see the Arduino web site 
///  and read about direct port I/O.
///  Once that is taken care of, the timer 0 compare A register 
///  is set to cause an interrupt at a count of 0x20 and then 
///  the interrupt is enabled.  As soon as it is enabled, the 
///  timer will begin interrupting and our ISR will start running
///  about 997 times per second.
////////////////////////////////////////////////////////////////

void stepper_initialize()
{
  // Initialize the output ports used
  DDRD |= 0xf0;     // D4 - D7 as outputs
  DDRB |= 0x0f;     // B0 - B3 as outputs
  DDRC |= 0x0f;     // C0 - C3 as outputs
  PORTD &= 0x0f;    // clear motor bits
  PORTB &= 0xf0;    // clear motor bits
  PORTC &= 0xf0;    // clear motor bits

  // Set up the timer interrupt
  // Set Output Compare Register 0A to a value away from 0.
  // We choose 0x20 (32 decimal) which should be plenty of time
  // and is also convenient if we want to add more interrupts later
  
  OCR0A = 0x20;  // Se the compare register away from 0
  
  // This line enables the interrupt for Output Compare Register 0A
  
  TIMSK0 |= 2;  // enable the compare A interrupt
}

///////////////////////////////////////////////////////////////////
///  stepper_setSteps
///  This function is perhaps the most important interface function. 
///  The motor given by "mot" is set to run for a certain number of 
///  steps with a given step size.  The step size (siz) can be 
///  negative to make the motor run backward.  A value of -5000 will 
///  step at the same rate as 5000 but in the opposite direction.
///////////////////////////////////////////////////////////////////

void stepper_setSteps(uint16_t mot, uint16_t steps, int16_t siz)
{
  if(mot < NUM_MOTORS)    // Make sure the motor number is valid.
  {
    noInterrupts();       // Don't allow interrupts while we make changes.
    stepper_steps[mot] = steps;
    stepper_stepSize[mot] = siz;
    interrupts();         // Done.  Allow interrupts now.
  }
}


//////////////////////////////////////////////////////////////////
///  stepper_stop
///  An individual motor can be stopped by calling this function.
///  Simply pass in the motor number and its steps will be set to 
///  zero.
//////////////////////////////////////////////////////////////////

void stepper_stop(uint16_t mot)
{
  if(mot < NUM_MOTORS)   // Make sure the motor number is valid.
  {
    noInterrupts();      // Don't allow interrupts while we make changes.
    stepper_steps[mot] = 0;
    interrupts();        // Done.  Allow interrupts now.
  }
}

//////////////////////////////////////////////////////////////////
///  stepper_startAll
///  All the motors that have been configured with stepsize and 
///  count will be started by this function.
//////////////////////////////////////////////////////////////////

void stepper_startAll()
{
  noInterrupts();      // Don't allow interrutps while we make changes.
  stepper_running = true;
  interrupts();        // Done.  Allow interrupts now.
}

//////////////////////////////////////////////////////////////////
///  stepper_stopAll
///  To stop all motors at once, such as in an emergency, call this
///  function.  All motors will have their steps remaining count 
///  set to zero.
///////////////////////////////////////////////////////////////////

void stepper_stopAll()
{
  noInterrupts();      // Don't allow interrupts while we make changes.
  stepper_running = false;
  for(int i = 0; i < NUM_MOTORS; i++)
  {
    stepper_steps[i] = 0;
  }
  interrupts();        // Done.  Allow interrupts now.
}

///////////////////////////////////////////////////////////////////
///  stepper_stepsRemaining
///  This function will let you know how many steps are remaining 
///  for any given motor.
///////////////////////////////////////////////////////////////////

uint16_t stepper_stepsRemaining(uint16_t mot)
{
  uint16_t rtn = 0;
  if(mot < NUM_MOTORS)
  {
    noInterrupts();    // Don't allow interrupts while we read.
    rtn = stepper_steps[mot];
    interrupts();      // Done.  Allow interrupts now.
  }
  return rtn;
}

//////////////////////////////////////////////////////////////////////////////
///  Interrupt Service Routine
///  ISR Timer 0, compare A interrupt
///  This is the interrupt service routine (ISR), the heart of the system.
///  Each time the timer 0 compare A register matches the timer 0 count, an 
///  interrupt is triggered and this function runs.
///  It starts by using the current phase for each motor to look up the correct 
///  output pattern and remembering that value.  It compares that 
///  output value to the previous output value.  If it is different that means 
///  a step has been taken so the steps remaining count is decremented.   Then 
///  the phase increment is added for each motor for the next interrupt.  Once 
///  all the calculations are done, the new values are output to the motors.
//////////////////////////////////////////////////////////////////////////////

// TODO:  If your Arduino uses something other than an ATMega328 
// you will need to change the TIMER0_COMPA_vect to the correct name

//  The Artduino (gcc) compiler has a special macro to 
//  declare an Insterrupt Service Routine.  The macro "ISR" takes a 
//  parameter with the name of the proper interrupt from:
//  https://www.nongnu.org/avr-libc/user-manual/group__avr__interrupts.html


ISR(TIMER0_COMPA_vect)
{
  static uint16_t oldIndex[NUM_MOTORS]; // store old index to test if we step

  if(stepper_running)
  {
    uint16_t index;
    uint8_t out[NUM_MOTORS];

    for(int mtr = 0; mtr < NUM_MOTORS; mtr++)
    {
      if(stepper_steps[mtr] != 0)
      {
        stepper_phase[mtr] += (uint16_t)stepper_stepSize[mtr];
        index = stepper_phase[mtr] >> stepper_indexShift;
        out[mtr] = stepper_stepTable[index];
        if(index != oldIndex[mtr])
        {
          stepper_steps[mtr]--;
          oldIndex[mtr] = index;
        }
      }
    }

    // The code below outputs the calculated values to the motors
    // It is specific to the hardware shown in the article.  For 
    // any other configuration / hardware you will need to write 
    // code to output the values to your motor controller.  It is 
    // HIGHLY recommended to use direct port I/O (as here) instead 
    // of Arduino digitalWrite() functions.  digitalWrite is slow
    // which will not only slow down your program, it will also 
    // cause the control bits to be written at different times 
    // which may cause performance issues and excess power 
    // consumption.
    uint8_t mot0 = PORTB & 0xf0;  // Get current value of port B.
    uint8_t mot1 = PORTD & 0x0f;  // Get current value of port D.
    uint8_t mot2 = PORTC & 0xf0;  // Get current value of port C.
    mot0 |= out[0];               // OR step bits into port B value.
    mot1 |= out[1] << 4;          // OR step bits into port D value.
    mot2 |= out[2];               // OR step bits into port C value.
    PORTB = mot0;            // Output combined bits to port B.
    PORTD = mot1;            // Output combined bits to port D.
    PORTC = mot2;            // OUtput combined bits to port C.
  }
}
